@@ -1,7 +1,7 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2010-2011 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2010-2015 -- leonerd@leonerd.org.uk
 
 package Net::Async::Tangence::Client;
 
@@ -10,10 +10,12 @@ use warnings;
 
 use base qw( Net::Async::Tangence::Protocol Tangence::Client );
 
-our $VERSION = '0.12';
+our $VERSION = '0.13';
 
 use Carp;
 
+use Future;
+
 use URI::Split qw( uri_split );
 
 =head1 NAME
@@ -105,22 +107,28 @@ sub configure
 
 =head1 METHODS
 
+The following methods documented with a trailing call to C<< ->get >> return
+L<Future> instances.
+
 =cut
 
-=head2 $client->connect_url( $url, %args )
+sub new_future
+{
+   my $self = shift;
+   return $self->loop->new_future;
+}
 
-Connects to a C<Tangence> server at the given URL.
+=head2 connect_url
+
+   $rootobj = $client->connect_url( $url, %args )->get
+
+Connects to a C<Tangence> server at the given URL. The returned L<Future> will
+yield the root object proxy once it has been obtained.
 
 Takes the following named arguments:
 
 =over 8
 
-=item on_connected => CODE
-
-Invoked once the connection to the server has been established.
-
- $on_connected->( $client )
-
 =item on_registry => CODE
 
 =item on_root => CODE
@@ -150,44 +158,58 @@ sub connect_url
       $scheme =~ s/^circle\+// or croak "Found a + within URL scheme that is not 'circle+'";
    }
 
-   my $on_connected = delete $args{on_connected};
-   $args{on_connected} = sub {
-      $on_connected->(@_) if $on_connected;
-      $self->tangence_connected( %args );
-   };
-
    # Legacy name
    $scheme = "sshexec" if $scheme eq "ssh";
 
+   my $f;
+
    if( $scheme eq "exec" ) {
       # Path will start with a leading /; we need to trim that
       $path =~ s{^/}{};
       # $query will contain args to exec - split them on +
       my @argv = split( m/\+/, $query );
-      return $self->connect_exec( [ $path, @argv ], %args );
+      $f = $self->connect_exec( [ $path, @argv ], %args );
    }
    elsif( $scheme eq "sshexec" ) {
       # Path will start with a leading /; we need to trim that
       $path =~ s{^/}{};
       # $query will contain args to exec - split them on +
       my @argv = split( m/\+/, $query );
-      return $self->connect_sshexec( $authority, [ $path, @argv ], %args );
+      $f = $self->connect_sshexec( $authority, [ $path, @argv ], %args );
    }
    elsif( $scheme eq "tcp" ) {
-      return $self->connect_tcp( $authority, %args );
+      $f = $self->connect_tcp( $authority, %args );
    }
    elsif( $scheme eq "unix" ) {
       # Path will start with a leading /; we need to trim that
       $path =~ s{^/}{};
-      return $self->connect_unix( $path, %args );
+      $f = $self->connect_unix( $path, %args );
    }
    elsif( $scheme eq "sshunix" ) {
       # Path will start with a leading /; we need to trim that
       $path =~ s{^/}{};
-      return $self->connect_sshunix( $authority, $path, %args );
+      $f = $self->connect_sshunix( $authority, $path, %args );
+   }
+   else {
+      croak "Unrecognised URL scheme name '$scheme'";
    }
 
-   croak "Unrecognised URL scheme name '$scheme'";
+   return $f->then( sub {
+      my $on_root = $args{on_root};
+
+      my $root_f = $self->new_future;
+
+      $self->tangence_connected( %args,
+         on_root => sub {
+            my ( $root ) = @_;
+
+            $on_root->( $root ) if $on_root;
+            $root_f->done( $root );
+         },
+      );
+
+      $root_f;
+   });
 }
 
 =item * exec
@@ -232,7 +254,7 @@ sub connect_exec
       write_handle => $mywrite,
    );
 
-   $args{on_connected}->( $self );
+   Future->done;
 }
 
 =item * sshexec
@@ -278,15 +300,6 @@ sub connect_tcp
    $self->connect(
       host     => $host,
       service  => $port,
-
-      on_connected => sub {
-         my ( $self ) = @_;
-
-         $args{on_connected}->( $self );
-      },
-
-      on_connect_error => sub { print STDERR "Cannot connect\n"; },
-      on_resolve_error => sub { print STDERR "Cannot resolve - $_[0]\n"; },
    );
 }
 
@@ -312,14 +325,6 @@ sub connect_unix
          socktype => 'stream',
          path     => $path,
       },
-
-      on_connected => sub {
-         my ( $self ) = @_;
-
-         $args{on_connected}->( $self );
-      },
-
-      on_connect_error => sub { print STDERR "Cannot connect\n"; },
    );
 }
 
@@ -368,23 +373,22 @@ sub connect_sshunix
    my $self = shift;
    my ( $host, $path, %args ) = @_;
 
-   my $on_connected = delete $args{on_connected};
-
    # Tell perl we're going to send it a program on STDIN
-   $self->connect_sshexec( $host, [ 'perl', '-', $path ],
-      %args,
-      on_connected => sub {
-         my $self = shift;
-         $self->write( _NC_MICRO . "\n__END__\n" );
-         $self->configure( on_read => sub {
-            my ( $self, $buffref, $eof ) = @_;
-            return 0 unless $$buffref =~ s/READY//;
-            $self->configure( on_read => undef );
-            $on_connected->( $self );
-            return 0;
-         } );
-      }
-   );
+   $self->connect_sshexec( $host, [ 'perl', '-', $path ], %args )
+   ->then( sub {
+      $self->write( _NC_MICRO . "\n__END__\n" );
+      my $f = $self->new_future;
+
+      $self->configure( on_read => sub {
+         my ( $self, $buffref, $eof ) = @_;
+         return 0 unless $$buffref =~ s/READY//;
+         $self->configure( on_read => undef );
+         $f->done;
+         return 0;
+      } );
+
+      return $f;
+   });
 }
 
 =back
